@@ -2,6 +2,7 @@
 import asyncio
 import copy
 import datetime
+import hashlib
 import logging
 import mimetypes
 import time
@@ -11,27 +12,24 @@ import uuid
 import warnings
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional, Union
+from typing import Any, Callable, Dict, List, Optional
 
 import requests
 
 from annofabapi import AnnofabApi
-from annofabapi.api import my_backoff
-from annofabapi.exceptions import AnnofabApiException
+from annofabapi.exceptions import AnnofabApiException, CheckSumError
 from annofabapi.models import (
     AdditionalData,
     AdditionalDataDefinitionType,
     AdditionalDataDefinitionV1,
     AnnotationDataHoldingType,
     AnnotationDetail,
-    AnnotationSpecsV1,
     FullAnnotationData,
     InputData,
     Inspection,
     InspectionStatus,
     Instruction,
     JobStatus,
-    JobType,
     LabelV1,
     MyOrganization,
     Organization,
@@ -46,7 +44,7 @@ from annofabapi.models import (
     TaskStatus,
 )
 from annofabapi.parser import SimpleAnnotationDirParser, SimpleAnnotationParser
-from annofabapi.utils import _download, _log_error_response, _raise_for_status, allow_404_error, str_now
+from annofabapi.utils import _download, _log_error_response, _raise_for_status, allow_404_error, my_backoff, str_now
 
 logger = logging.getLogger(__name__)
 
@@ -216,6 +214,28 @@ class Wrapper:
 
         return all_objects
 
+    @my_backoff
+    def _request_get_wrapper(self, url: str) -> requests.Response:
+        """
+        HTTP GETのリクエスト。
+        リトライするためにメソッドを切り出した。
+        """
+        return self.api.session.get(url)
+
+    @my_backoff
+    def _request_put_wrapper(
+        self,
+        url: str,
+        params: Optional[Dict[str, Any]] = None,
+        data: Optional[Any] = None,
+        headers: Optional[Dict[str, Any]] = None,
+    ) -> requests.Response:
+        """
+        HTTP PUTのリクエスト。
+        リトライするためにメソッドを切り出した
+        """
+        return self.api.session.put(url, params=params, data=data, headers=headers)
+
     #########################################
     # Public Method : Annotation
     #########################################
@@ -327,23 +347,38 @@ class Wrapper:
     ) -> Dict[str, Any]:
         """
         コピー元の１個のアノテーションを、コピー先用に変換する。
-        塗りつぶし画像の場合、S3にアップロードする。
+        塗りつぶし画像などの外部アノテーションファイルがある場合、S3にアップロードする。
 
         Notes:
             annotation_id をUUIDv4で生成すると、アノテーションリンク属性をコピーしたときに対応できないので、暫定的にannotation_idは維持するようにする。
+
+        Raises:
+            CheckSumError: アップロードした外部アノテーションファイルのMD5ハッシュ値が、S3にアップロードしたときのレスポンスのETagに一致しない
+
         """
         dest_detail = detail
         dest_detail["account_id"] = account_id
         if detail["data_holding_type"] == AnnotationDataHoldingType.OUTER.value:
-            outer_file_url = detail["url"]
-            src_response = self.api.session.get(outer_file_url)
-            s3_path = self.upload_data_to_s3(
-                dest_project_id, data=src_response.content, content_type=src_response.headers["Content-Type"]
-            )
-            logger.debug("%s に塗りつぶし画像をアップロードしました。", s3_path)
-            dest_detail["path"] = s3_path
-            dest_detail["url"] = None
-            dest_detail["etag"] = None
+
+            try:
+                outer_file_url = detail["url"]
+                src_response = self._request_get_wrapper(outer_file_url)
+                s3_path = self.upload_data_to_s3(
+                    dest_project_id, data=src_response.content, content_type=src_response.headers["Content-Type"]
+                )
+                logger.debug("%s に外部アノテーションファイルをアップロードしました。", s3_path)
+                dest_detail["path"] = s3_path
+                dest_detail["url"] = None
+                dest_detail["etag"] = None
+
+            except CheckSumError as e:
+                message = (
+                    f"外部アノテーションファイル {outer_file_url} のレスポンスのMD5ハッシュ値('{e.uploaded_data_hash}')が、"
+                    f"AWS S3にアップロードしたときのレスポンスのETag('{e.response_etag}')に一致しませんでした。アップロード時にデータが破損した可能性があります。"
+                )
+                raise CheckSumError(
+                    message=message, uploaded_data_hash=e.uploaded_data_hash, response_etag=e.response_etag
+                ) from e
 
         return dest_detail
 
@@ -514,6 +549,9 @@ class Wrapper:
 
         Returns:
 
+        Raises:
+            CheckSumError: アップロードした外部アノテーションファイルのMD5ハッシュ値が、S3にアップロードしたときのレスポンスのETagに一致しない
+
         """
         label_info = self.__get_label_info_from_label_name(detail["label"], annotation_specs_labels)
         if label_info is None:
@@ -540,10 +578,21 @@ class Wrapper:
 
         if data_holding_type == AnnotationDataHoldingType.OUTER.value:
             data_uri = detail["data"]["data_uri"]
+            outer_file_path = f"{parser.task_id}/{parser.input_data_id}/{data_uri}"
             with parser.open_outer_file(data_uri) as f:
-                s3_path = self.upload_data_to_s3(project_id, f, content_type="image/png")
-                dest_obj["path"] = s3_path
-                logger.debug(f"{parser.task_id}/{parser.input_data_id}/{data_uri} をS3にアップロードしました。")
+                try:
+                    s3_path = self.upload_data_to_s3(project_id, f, content_type="image/png")
+                    dest_obj["path"] = s3_path
+                    logger.debug(f"{outer_file_path} をS3にアップロードしました。")
+
+                except CheckSumError as e:
+                    message = (
+                        f"アップロードした外部アノテーションファイル'{outer_file_path}'のMD5ハッシュ値('{e.uploaded_data_hash}')が、"
+                        f"AWS S3にアップロードしたときのレスポンスのETag('{e.response_etag}')に一致しませんでした。アップロード時にデータが破損した可能性があります。"
+                    )
+                    raise CheckSumError(
+                        message=message, uploaded_data_hash=e.uploaded_data_hash, response_etag=e.response_etag
+                    ) from e
 
         return dest_obj
 
@@ -648,39 +697,6 @@ class Wrapper:
     #########################################
     # Public Method : AnnotationSpecs
     #########################################
-    def copy_annotation_specs(
-        self, src_project_id: str, dest_project_id: str, comment: Optional[str] = None
-    ) -> AnnotationSpecsV1:
-        """
-        アノテーション仕様を、別のプロジェクトにコピーする。
-
-        .. deprecated:: 2021-09-01
-
-
-        Note:
-            誤って実行しないようにすること
-
-        Args:
-            src_project_id: コピー元のproject_id
-            dest_project_id: コピー先のproject_id
-            comment: アノテーション仕様を保存するときのコメント。Noneならば、コピーした旨を記載する。
-
-        Returns:
-            put_annotation_specsのContent
-        """
-        warnings.warn("2021-09-01以降に削除します。", FutureWarning)
-
-        src_annotation_specs = self.api.get_annotation_specs(src_project_id)[0]
-
-        if comment is None:
-            comment = f"Copied the annotation specification of project {src_project_id} on {str_now()}"
-
-        request_body = {
-            "labels": src_annotation_specs["labels"],
-            "inspection_phrases": src_annotation_specs["inspection_phrases"],
-            "comment": comment,
-        }
-        return self.api.put_annotation_specs(dest_project_id, request_body=request_body)[0]
 
     @staticmethod
     def __get_label_name_en(label: Dict[str, Any]) -> str:
@@ -844,12 +860,25 @@ class Wrapper:
 
         Returns:
             一時データ保存先であるS3パス
+
+        Raises:
+            CheckSumError: アップロードしたファイルのMD5ハッシュ値が、S3にアップロードしたときのレスポンスのETagと一致しない
+
         """
 
         # content_type を推測
         new_content_type = self._get_content_type(file_path, content_type)
         with open(file_path, "rb") as f:
-            return self.upload_data_to_s3(project_id, data=f, content_type=new_content_type)
+            try:
+                return self.upload_data_to_s3(project_id, data=f, content_type=new_content_type)
+            except CheckSumError as e:
+                message = (
+                    f"アップロードしたファイル'{file_path}'のMD5ハッシュ値('{e.uploaded_data_hash}')が、"
+                    f"AWS S3にアップロードしたときのレスポンスのETag('{e.response_etag}')に一致しませんでした。アップロード時にデータが破損した可能性があります。"
+                )
+                raise CheckSumError(
+                    message=message, uploaded_data_hash=e.uploaded_data_hash, response_etag=e.response_etag
+                ) from e
 
     def upload_data_to_s3(self, project_id: str, data: Any, content_type: str) -> str:
         """
@@ -857,12 +886,25 @@ class Wrapper:
 
         Args:
             project_id: プロジェクトID
-            data: アップロード対象のdata. ``requests.put`` メソッドの ``data`` 引数にそのまま渡す。
+            data: アップロード対象のdata. ``open(mode="b")`` 関数の戻り値、またはバイナリ型の値です。 ``requests.put`` メソッドの ``data`` 引数にそのまま渡します。
             content_type: アップロードするfile objectのMIME Type.
 
         Returns:
             一時データ保存先であるS3パス
+
+        Raises:
+            CheckSumError: アップロードしたデータのMD5ハッシュ値が、S3にアップロードしたときのレスポンスのETagと一致しない
         """
+
+        def get_md5_value_from_file(fp):
+            md5_obj = hashlib.md5()
+            while True:
+                chunk = fp.read(2048 * md5_obj.block_size)
+                if len(chunk) == 0:
+                    break
+                md5_obj.update(chunk)
+            return md5_obj.hexdigest()
+
         # 一時データ保存先を取得
         content = self.api.create_temp_path(project_id, header_params={"content-type": content_type})[0]
 
@@ -873,10 +915,31 @@ class Wrapper:
         s3_url = content["url"].split("?")[0]
 
         # アップロード
-        res_put = self.api.session.put(s3_url, params=query_dict, data=data, headers={"content-type": content_type})
+        res_put = self._request_put_wrapper(
+            url=s3_url, params=query_dict, data=data, headers={"content-type": content_type}
+        )
 
         _log_error_response(logger, res_put)
         _raise_for_status(res_put)
+
+        # アップロードしたファイルが破損していなかをチェックする
+        if hasattr(data, "read"):
+            # 読み込み位置を先頭に戻す
+            data.seek(0)
+            uploaded_data_hash = get_md5_value_from_file(data)
+        else:
+            uploaded_data_hash = hashlib.md5(data).hexdigest()
+
+        # ETagにはダブルクォートが含まれているため、`str_md5`もそれに合わせる
+        response_etag = res_put.headers["ETag"]
+
+        if f'"{uploaded_data_hash}"' != response_etag:
+            message = (
+                f"アップロードしたデータのMD5ハッシュ値('{uploaded_data_hash}')が、"
+                f"AWS S3にアップロードしたときのレスポンスのETag('{response_etag}')に一致しませんでした。アップロード時にデータが破損した可能性があります。"
+            )
+            raise CheckSumError(message=message, uploaded_data_hash=uploaded_data_hash, response_etag=response_etag)
+
         return content["path"]
 
     def put_input_data_from_file(
@@ -935,7 +998,7 @@ class Wrapper:
             logger.warning(f"レスポンスヘッダに'Location'がありません。method={response.request.method}, url={response.request.url}")
             return None
 
-        response = self.api.session.get(url)
+        response = self._request_get_wrapper(url)
         _log_error_response(logger, response)
 
         response.encoding = "utf-8"
@@ -1720,7 +1783,9 @@ class Wrapper:
         s3_url = content["url"].split("?")[0]
 
         # アップロード
-        res_put = self.api.session.put(s3_url, params=query_dict, data=data, headers={"content-type": content_type})
+        res_put = self._request_put_wrapper(
+            url=s3_url, params=query_dict, data=data, headers={"content-type": content_type}
+        )
         _log_error_response(logger, res_put)
         _raise_for_status(res_put)
         return content["path"]
@@ -1728,9 +1793,7 @@ class Wrapper:
     #########################################
     # Public Method : Job
     #########################################
-    def delete_all_succeeded_job(
-        self, project_id: str, job_type: Union[ProjectJobType, JobType]
-    ) -> List[ProjectJobInfo]:
+    def delete_all_succeeded_job(self, project_id: str, job_type: ProjectJobType) -> List[ProjectJobInfo]:
         """
         成功したジョブをすべて削除する
 
@@ -1772,7 +1835,7 @@ class Wrapper:
         all_jobs.extend(r["list"])
         return all_jobs
 
-    def job_in_progress(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
+    def job_in_progress(self, project_id: str, job_type: ProjectJobType) -> bool:
         """
         ジョブが進行中かどうか
 
@@ -1794,7 +1857,7 @@ class Wrapper:
     def wait_for_completion(
         self,
         project_id: str,
-        job_type: Union[ProjectJobType, JobType],
+        job_type: ProjectJobType,
         job_access_interval: int = 60,
         max_job_access: int = 10,
     ) -> bool:
@@ -1824,7 +1887,7 @@ class Wrapper:
     def wait_until_job_finished(
         self,
         project_id: str,
-        job_type: Union[ProjectJobType, JobType],
+        job_type: ProjectJobType,
         job_id: Optional[str] = None,
         job_access_interval: int = 60,
         max_job_access: int = 360,
@@ -1902,14 +1965,14 @@ class Wrapper:
                     )
                     return JobStatus.PROGRESS
 
-    async def _job_in_progress_async(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
+    async def _job_in_progress_async(self, project_id: str, job_type: ProjectJobType) -> bool:
         loop = asyncio.get_event_loop()
         return await loop.run_in_executor(None, self.job_in_progress, project_id, job_type)
 
     async def _wait_until_job_finished_async(
         self,
         project_id: str,
-        job_type: Union[ProjectJobType, JobType],
+        job_type: ProjectJobType,
         job_id: Optional[str],
         job_access_interval: int,
         max_job_access: int,
@@ -1919,7 +1982,7 @@ class Wrapper:
             None, self.wait_until_job_finished, project_id, job_type, job_id, job_access_interval, max_job_access
         )
 
-    def can_execute_job(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
+    def can_execute_job(self, project_id: str, job_type: ProjectJobType) -> bool:
         """
         ジョブが実行できる状態か否か。他のジョブが実行中で同時に実行できない場合はFalseを返す。
 
@@ -1930,19 +1993,14 @@ class Wrapper:
         Returns:
             ジョブが実行できる状態か否か
         """
-        # TODO: JobTypeが削除されたら、この処理も削除する
-        new_job_type: ProjectJobType = ProjectJobType(job_type.value) if isinstance(job_type, JobType) else job_type
-
-        job_type_list = _JOB_CONCURRENCY_LIMIT[new_job_type]
+        job_type_list = _JOB_CONCURRENCY_LIMIT[job_type]
 
         # tokenがない場合、ログインが複数回発生するので、事前にログインしておく
         if self.api.token_dict is None:
             self.api.login()
 
         # 複数のジョブに対して進行中かどうかを確認する
-        gather = asyncio.gather(
-            *[self._job_in_progress_async(project_id, new_job_type) for new_job_type in job_type_list]
-        )
+        gather = asyncio.gather(*[self._job_in_progress_async(project_id, job_type) for job_type in job_type_list])
         loop = asyncio.get_event_loop()
         result = loop.run_until_complete(gather)
 
@@ -1951,7 +2009,7 @@ class Wrapper:
     def wait_until_job_is_executable(
         self,
         project_id: str,
-        job_type: Union[ProjectJobType, JobType],
+        job_type: ProjectJobType,
         job_access_interval: int = 60,
         max_job_access: int = 360,
     ) -> bool:
@@ -1968,10 +2026,8 @@ class Wrapper:
             指定した時間（アクセス頻度と回数）待った後、ジョブが実行可能な状態かどうか。進行中のジョブが存在する場合は、ジョブが実行不可能。
 
         """
-        # TODO: JobTypeが削除されたら、この処理も削除する
-        new_job_type: ProjectJobType = ProjectJobType(job_type.value) if isinstance(job_type, JobType) else job_type
 
-        job_type_list = _JOB_CONCURRENCY_LIMIT[new_job_type]
+        job_type_list = _JOB_CONCURRENCY_LIMIT[job_type]
         # tokenがない場合、ログインが複数回発生するので、事前にログインしておく
         if self.api.token_dict is None:
             self.api.login()
